@@ -34,6 +34,7 @@
 #include <errno.h>
 
 #include "utils.h"
+#include "nm-utils/nm-shared-utils.h"
 
 #define BLOCK_HANDLER_ID "block-handler-id"
 
@@ -68,6 +69,25 @@ setup_secret_widget (GtkBuilder *builder,
 	return widget;
 }
 
+/* From gnome-control-center/panels/network/connection-editor/ui-helpers.c */
+
+static void
+widget_set_error (GtkWidget *widget)
+{
+	g_return_if_fail (GTK_IS_WIDGET (widget));
+
+	gtk_style_context_add_class (gtk_widget_get_style_context (widget), "error");
+}
+
+static void
+widget_unset_error (GtkWidget *widget)
+{
+	g_return_if_fail (GTK_IS_WIDGET (widget));
+
+	gtk_style_context_remove_class (gtk_widget_get_style_context (widget), "error");
+}
+
+
 typedef struct {
 	GtkWidget *widget1;
 	GtkWidget *widget2;
@@ -156,7 +176,7 @@ tls_setup (GtkBuilder *builder,
 	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (cert), filter);
 	gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (cert), TRUE);
 	gtk_file_chooser_button_set_title (GTK_FILE_CHOOSER_BUTTON (cert),
-	                                   _("Choose your personal certificate..."));
+	                                   _("Choose your personal certificate…"));
 	g_signal_connect (G_OBJECT (cert), "selection-changed", G_CALLBACK (changed_cb), user_data);
 
 	if (s_vpn) {
@@ -174,7 +194,7 @@ tls_setup (GtkBuilder *builder,
 	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (key), filter);
 	gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (key), TRUE);
 	gtk_file_chooser_button_set_title (GTK_FILE_CHOOSER_BUTTON (key),
-	                                   _("Choose your private key..."));
+	                                   _("Choose your private key…"));
 	g_signal_connect (G_OBJECT (key), "selection-changed", G_CALLBACK (changed_cb), user_data);
 
 	if (s_vpn) {
@@ -290,7 +310,7 @@ tls_pw_init_auth_widget (GtkBuilder *builder,
 	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (ca), filter);
 	gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (ca), TRUE);
 	gtk_file_chooser_button_set_title (GTK_FILE_CHOOSER_BUTTON (ca),
-	                                   _("Choose a Certificate Authority certificate..."));
+	                                   _("Choose a Certificate Authority certificate…"));
 	g_signal_connect (G_OBJECT (ca), "selection-changed", G_CALLBACK (changed_cb), user_data);
 
 	if (s_vpn) {
@@ -334,7 +354,7 @@ sk_init_auth_widget (GtkBuilder *builder,
 	gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (widget), filter);
 	gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (widget), TRUE);
 	gtk_file_chooser_button_set_title (GTK_FILE_CHOOSER_BUTTON (widget),
-	                                   _("Choose an OpenVPN static key..."));
+	                                   _("Choose an OpenVPN static key…"));
 	g_signal_connect (G_OBJECT (widget), "selection-changed", G_CALLBACK (changed_cb), user_data);
 
 	if (s_vpn) {
@@ -888,6 +908,7 @@ static const char *advanced_keys[] = {
 	NM_OPENVPN_KEY_TA,
 	NM_OPENVPN_KEY_RENEG_SECONDS,
 	NM_OPENVPN_KEY_TLS_REMOTE,
+	NM_OPENVPN_KEY_VERIFY_X509_NAME,
 	NM_OPENVPN_KEY_REMOTE_RANDOM,
 	NM_OPENVPN_KEY_TUN_IPV6,
 	NM_OPENVPN_KEY_REMOTE_CERT_TLS,
@@ -895,6 +916,7 @@ static const char *advanced_keys[] = {
 	NM_OPENVPN_KEY_PING,
 	NM_OPENVPN_KEY_PING_EXIT,
 	NM_OPENVPN_KEY_PING_RESTART,
+	NM_OPENVPN_KEY_MAX_ROUTES,
 	NULL
 };
 
@@ -981,7 +1003,7 @@ populate_cipher_combo (GtkComboBox *box, const char *user_cipher)
 	gboolean user_added = FALSE;
 	char *argv[3];
 	GError *error = NULL;
-	gboolean success, found_blank = FALSE;
+	gboolean success, ignore_lines = TRUE;
 
 	openvpn_binary = nm_find_openvpn ();
 	if (!openvpn_binary)
@@ -1020,15 +1042,21 @@ populate_cipher_combo (GtkComboBox *box, const char *user_cipher)
 	g_free (tmp);
 
 	for (item = items; *item; item++) {
-		char *space = strchr (*item, ' ');
+		char *space;
 
-		/* Don't add anything until after the first blank line */
-		if (!found_blank) {
-			if (!strlen (*item))
-				found_blank = TRUE;
+		/* Don't add anything until after the first blank line. Also,
+		 * any blank line indicates the start of a comment, ended by
+		 * another blank line.
+		 */
+		if (!strlen (*item)) {
+			ignore_lines = !ignore_lines;
 			continue;
 		}
 
+		if (ignore_lines)
+			continue;
+
+		space = strchr (*item, ' ');
 		if (space)
 			*space = '\0';
 
@@ -1040,7 +1068,7 @@ populate_cipher_combo (GtkComboBox *box, const char *user_cipher)
 			gtk_list_store_set (store, &iter,
 			                    TLS_CIPHER_COL_NAME, *item,
 			                    TLS_CIPHER_COL_DEFAULT, FALSE, -1);
-			if (!user_added && user_cipher && !strcmp (*item, user_cipher)) {
+			if (!user_added && user_cipher && !g_ascii_strcasecmp (*item, user_cipher)) {
 				gtk_combo_box_set_active_iter (box, &iter);
 				user_added = TRUE;
 			}
@@ -1052,7 +1080,7 @@ populate_cipher_combo (GtkComboBox *box, const char *user_cipher)
 		gtk_list_store_insert (store, &iter, 1);
 		gtk_list_store_set (store, &iter,
 		                    TLS_CIPHER_COL_NAME, user_cipher,
-		                    TLS_CIPHER_COL_DEFAULT, FALSE -1);
+		                    TLS_CIPHER_COL_DEFAULT, FALSE, -1);
 		gtk_combo_box_set_active_iter (box, &iter);
 	} else if (!user_added) {
 		gtk_combo_box_set_active (box, 0);
@@ -1125,7 +1153,7 @@ populate_hmacauth_combo (GtkComboBox *box, const char *hmacauth)
 		                    HMACAUTH_COL_NAME, name,
 		                    HMACAUTH_COL_VALUE, *item,
 		                    HMACAUTH_COL_DEFAULT, FALSE, -1);
-		if (hmacauth && !strcmp (*item, hmacauth)) {
+		if (hmacauth && !g_ascii_strcasecmp (*item, hmacauth)) {
 			gtk_combo_box_set_active_iter (box, &iter);
 			active_initialized = TRUE;
 		}
@@ -1137,6 +1165,126 @@ populate_hmacauth_combo (GtkComboBox *box, const char *hmacauth)
 	g_object_unref (store);
 }
 
+#define TLS_REMOTE_MODE_NONE        "none"
+#define TLS_REMOTE_MODE_SUBJECT     NM_OPENVPN_VERIFY_X509_NAME_TYPE_SUBJECT
+#define TLS_REMOTE_MODE_NAME        NM_OPENVPN_VERIFY_X509_NAME_TYPE_NAME
+#define TLS_REMOTE_MODE_NAME_PREFIX NM_OPENVPN_VERIFY_X509_NAME_TYPE_NAME_PREFIX
+#define TLS_REMOTE_MODE_LEGACY      "legacy"
+
+#define TLS_REMOTE_MODE_COL_NAME 0
+#define TLS_REMOTE_MODE_COL_VALUE 1
+
+static void
+populate_tls_remote_mode_entry_combo (GtkEntry* entry, GtkComboBox *box,
+                                      const char *tls_remote, const char *x509_name)
+{
+	GtkListStore *store;
+	GtkTreeIter iter;
+	const char *subject_name = NULL;
+
+	store = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_STRING);
+	gtk_combo_box_set_model (box, GTK_TREE_MODEL (store));
+
+	gtk_list_store_append (store, &iter);
+	gtk_list_store_set (store, &iter,
+	                    TLS_REMOTE_MODE_COL_NAME, _("Don’t verify certificate identification"),
+	                    TLS_REMOTE_MODE_COL_VALUE, TLS_REMOTE_MODE_NONE,
+	                    -1);
+
+	gtk_list_store_append (store, &iter);
+	gtk_list_store_set (store, &iter,
+	                    TLS_REMOTE_MODE_COL_NAME, _("Verify whole subject exactly"),
+	                    TLS_REMOTE_MODE_COL_VALUE, TLS_REMOTE_MODE_SUBJECT,
+	                    -1);
+
+	gtk_list_store_append (store, &iter);
+	gtk_list_store_set (store, &iter,
+	                    TLS_REMOTE_MODE_COL_NAME, _("Verify name exactly"),
+	                    TLS_REMOTE_MODE_COL_VALUE, TLS_REMOTE_MODE_NAME,
+	                    -1);
+
+	gtk_list_store_append (store, &iter);
+	gtk_list_store_set (store, &iter,
+	                    TLS_REMOTE_MODE_COL_NAME, _("Verify name by prefix"),
+	                    TLS_REMOTE_MODE_COL_VALUE, TLS_REMOTE_MODE_NAME_PREFIX,
+	                    -1);
+
+	gtk_list_store_append (store, &iter);
+	gtk_list_store_set (store, &iter,
+	                    TLS_REMOTE_MODE_COL_NAME, _("Verify subject partially (legacy mode)"),
+	                    TLS_REMOTE_MODE_COL_VALUE, TLS_REMOTE_MODE_LEGACY,
+	                    -1);
+
+	if (x509_name && strlen (x509_name)) {
+		if (g_str_has_prefix (x509_name, "name:"))
+			gtk_combo_box_set_active (box, 2);
+		else if (g_str_has_prefix (x509_name, "name-prefix:"))
+			gtk_combo_box_set_active (box, 3);
+		else
+			gtk_combo_box_set_active (box, 1);
+
+		subject_name = strchr (x509_name, ':');
+		if (subject_name)
+			subject_name++;
+		else
+			subject_name = x509_name;
+	} else if (tls_remote && strlen (tls_remote)) {
+		gtk_combo_box_set_active (box, 4);
+
+		subject_name = tls_remote;
+	} else {
+		gtk_combo_box_set_active (box, 0);
+
+		subject_name = "";
+	}
+
+	gtk_entry_set_text (entry, subject_name);
+
+    g_object_unref (store);
+}
+
+static void
+tls_remote_changed (GtkWidget *widget, gpointer user_data)
+{
+	GtkBuilder *builder = (GtkBuilder *) user_data;
+	GtkWidget *entry, *combo, *ok_button;
+	GtkTreeIter iter;
+	gboolean entry_enabled = TRUE, entry_has_error = FALSE;
+
+	entry     = GTK_WIDGET (gtk_builder_get_object (builder, "tls_remote_entry"));
+	combo     = GTK_WIDGET (gtk_builder_get_object (builder, "tls_remote_mode_combo"));
+	ok_button = GTK_WIDGET (gtk_builder_get_object (builder, "ok_button"));
+
+	if (gtk_combo_box_get_active_iter (GTK_COMBO_BOX (combo), &iter)) {
+		gs_free char *tls_remote_mode = NULL;
+		GtkTreeModel *combo_model = gtk_combo_box_get_model (GTK_COMBO_BOX (combo));
+
+		gtk_tree_model_get (combo_model, &iter, TLS_REMOTE_MODE_COL_VALUE, &tls_remote_mode, -1);
+		g_return_if_fail (tls_remote_mode);
+
+		/* If a mode of 'none' is selected, disable the subject entry control.
+		   Otherwise, enable the entry, and set up it's error state based on
+		   whether it is empty or not (it should not be). */
+		if (!strcmp (tls_remote_mode, TLS_REMOTE_MODE_NONE)) {
+			entry_enabled = FALSE;
+		} else {
+			const char *subject = gtk_entry_get_text (GTK_ENTRY (entry));
+
+			entry_enabled = TRUE;
+			entry_has_error = !subject || !subject[0];
+		}
+	}
+
+	gtk_widget_set_sensitive (entry, entry_enabled);
+	if(entry_has_error) {
+		widget_set_error (entry);
+		gtk_widget_set_sensitive (ok_button, FALSE);
+	} else {
+		widget_unset_error (entry);
+		gtk_widget_set_sensitive (ok_button, TRUE);
+	}
+}
+
 static void
 remote_tls_cert_toggled_cb (GtkWidget *widget, gpointer user_data)
 {
@@ -1359,18 +1507,16 @@ device_name_changed_cb (GtkEntry *entry,
 	GtkWidget *ok_button = user_data;
 	gboolean entry_sensitive;
 	char *entry_text;
-	GdkRGBA rgba;
 
 	entry_sensitive = gtk_widget_get_sensitive (GTK_WIDGET (entry));
 	entry_text = gtk_editable_get_chars (editable, 0, -1);
 
 	/* Change cell's background to red if the value is invalid */
 	if (entry_sensitive && entry_text[0] != '\0' && !nm_utils_iface_valid_name (entry_text)) {
-		gdk_rgba_parse (&rgba, "red");
-		gtk_widget_override_background_color (GTK_WIDGET (editable), GTK_STATE_FLAG_NORMAL, &rgba);
+		widget_set_error (GTK_WIDGET (editable));
 		gtk_widget_set_sensitive (ok_button, FALSE);
 	} else {
-		gtk_widget_override_background_color (GTK_WIDGET (editable), GTK_STATE_FLAG_NORMAL, NULL);
+		widget_unset_error (GTK_WIDGET (editable));
 		gtk_widget_set_sensitive (ok_button, TRUE);
 	}
 
@@ -1399,6 +1545,57 @@ dev_checkbox_toggled_cb (GtkWidget *check, gpointer user_data)
 	device_name_changed_cb (GTK_ENTRY (entry), ok_button);
 }
 
+static gboolean
+_hash_get_boolean (GHashTable *hash,
+                   const char *key)
+{
+	const char *value;
+
+	nm_assert (hash);
+	nm_assert (key && key[0]);
+
+	value = g_hash_table_lookup (hash, key);
+
+	return nm_streq0 (value, "yes");
+}
+
+static void
+_builder_init_toggle_button (GtkBuilder *builder,
+                             const char *widget_name,
+                             gboolean active_state)
+{
+	GtkToggleButton *widget;
+
+	widget = (GtkToggleButton *) gtk_builder_get_object (builder, widget_name);
+	g_return_if_fail (GTK_IS_TOGGLE_BUTTON (widget));
+
+	gtk_toggle_button_set_active (widget, active_state);
+}
+
+static void
+_builder_init_optional_spinbutton (GtkBuilder *builder,
+                                   const char *checkbutton_name,
+                                   const char *spinbutton_name,
+                                   gboolean active_state,
+                                   gint64 value)
+{
+	GtkWidget *widget;
+	GtkWidget *spin;
+
+	widget = (GtkWidget *) gtk_builder_get_object (builder, checkbutton_name);
+	g_return_if_fail (GTK_IS_TOGGLE_BUTTON (widget));
+
+	spin = (GtkWidget *) gtk_builder_get_object (builder, spinbutton_name);
+	g_return_if_fail (GTK_IS_SPIN_BUTTON (spin));
+
+	g_signal_connect ((GObject *) widget, "toggled", G_CALLBACK (checkbox_toggled_update_widget_cb), spin);
+
+	gtk_spin_button_set_value ((GtkSpinButton *) spin, (double) value);
+
+	gtk_widget_set_sensitive (spin, active_state);
+	gtk_toggle_button_set_active ((GtkToggleButton *) widget, active_state);
+}
+
 static void
 ping_exit_restart_checkbox_toggled_cb (GtkWidget *check, gpointer user_data)
 {
@@ -1457,30 +1654,11 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 
 	ok_button = GTK_WIDGET (gtk_builder_get_object (builder, "ok_button"));
 
-	widget = GTK_WIDGET (gtk_builder_get_object (builder, "reneg_checkbutton"));
-	spin = GTK_WIDGET (gtk_builder_get_object (builder, "reneg_spinbutton"));
-	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (checkbox_toggled_update_widget_cb), spin);
 
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_RENEG_SECONDS);
-	if (value && strlen (value)) {
-		long int tmp;
+	_builder_init_optional_spinbutton (builder, "reneg_checkbutton", "reneg_spinbutton", !!value,
+	                                   _nm_utils_ascii_str_to_int64 (value, 10, 0, G_MAXINT, 0));
 
-		errno = 0;
-		tmp = strtol (value, NULL, 10);
-		if (errno == 0 && tmp >= 0 && tmp <= 604800) {  /* up to a week? */
-			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
-
-			widget = GTK_WIDGET (gtk_builder_get_object (builder, "reneg_spinbutton"));
-			gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), (gdouble) tmp);
-		}
-		gtk_widget_set_sensitive (widget, TRUE);
-	} else {
-		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), FALSE);
-
-		widget = GTK_WIDGET (gtk_builder_get_object (builder, "reneg_spinbutton"));
-		gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), 0.0);
-		gtk_widget_set_sensitive (widget, FALSE);
-	}
 
 	/* Proxy support */
 	combo = GTK_WIDGET (gtk_builder_get_object (builder, "proxy_type_combo"));
@@ -1556,111 +1734,34 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 	widget = GTK_WIDGET (gtk_builder_get_object (builder, "show_proxy_password"));
 	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (show_proxy_password_toggled_cb), builder);
 
-	widget = GTK_WIDGET (gtk_builder_get_object (builder, "port_checkbutton"));
-	spin = GTK_WIDGET (gtk_builder_get_object (builder, "port_spinbutton"));
-	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (checkbox_toggled_update_widget_cb), spin);
 
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_PORT);
-	if (value && strlen (value)) {
-		long int tmp;
+	_builder_init_optional_spinbutton (builder, "port_checkbutton", "port_spinbutton", !!value,
+	                                   _nm_utils_ascii_str_to_int64 (value, 10, 1, 65535, 1194));
 
-		errno = 0;
-		tmp = strtol (value, NULL, 10);
-		if (errno == 0 && tmp > 0 && tmp < 65536 && tmp != 1194) {
-			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
-
-			widget = GTK_WIDGET (gtk_builder_get_object (builder, "port_spinbutton"));
-			gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget),
-			                           (gdouble) tmp);
-		}
-		gtk_widget_set_sensitive (widget, TRUE);
-	} else {
-		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), FALSE);
-
-		widget = GTK_WIDGET (gtk_builder_get_object (builder, "port_spinbutton"));
-		gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), 1194.0);
-		gtk_widget_set_sensitive (widget, FALSE);
-	}
-
-	widget = GTK_WIDGET (gtk_builder_get_object (builder, "tunmtu_checkbutton"));
-	g_assert (widget);
-	spin = GTK_WIDGET (gtk_builder_get_object (builder, "tunmtu_spinbutton"));
-	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (checkbox_toggled_update_widget_cb), spin);
 
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_TUNNEL_MTU);
-	if (value && strlen (value)) {
-		long int tmp;
+	_builder_init_optional_spinbutton (builder, "tunmtu_checkbutton", "tunmtu_spinbutton", !!value,
+	                                   _nm_utils_ascii_str_to_int64 (value, 10, 1, 65535, 1500));
 
-		errno = 0;
-		tmp = strtol (value, NULL, 10);
-		if (errno == 0 && tmp > 0 && tmp < 65536) {
-			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
-
-			widget = GTK_WIDGET (gtk_builder_get_object (builder, "tunmtu_spinbutton"));
-			gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), (gdouble) tmp);
-			gtk_widget_set_sensitive (widget, TRUE);
-		}
-	} else {
-		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), FALSE);
-
-		widget = GTK_WIDGET (gtk_builder_get_object (builder, "tunmtu_spinbutton"));
-		gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), 1500.0);
-		gtk_widget_set_sensitive (widget, FALSE);
-	}
-
-	widget = GTK_WIDGET (gtk_builder_get_object (builder, "fragment_checkbutton"));
-	spin = GTK_WIDGET (gtk_builder_get_object (builder, "fragment_spinbutton"));
-	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (checkbox_toggled_update_widget_cb), spin);
 
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_FRAGMENT_SIZE);
-	if (value && strlen (value)) {
-		long int tmp;
+	_builder_init_optional_spinbutton (builder, "fragment_checkbutton", "fragment_spinbutton", !!value,
+	                                   _nm_utils_ascii_str_to_int64 (value, 10, 0, 65535, 1300));
 
-		errno = 0;
-		tmp = strtol (value, NULL, 10);
-		if (errno == 0 && tmp >= 0 && tmp < 65536) {
-			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
-
-			widget = GTK_WIDGET (gtk_builder_get_object (builder, "fragment_spinbutton"));
-			gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), (gdouble) tmp);
-			gtk_widget_set_sensitive (widget, TRUE);
-		}
-	} else {
-		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), FALSE);
-
-		widget = GTK_WIDGET (gtk_builder_get_object (builder, "fragment_spinbutton"));
-		gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), 1300.0);
-		gtk_widget_set_sensitive (widget, FALSE);
-	}
 
+	/* the UI currently only supports "--comp-lzo yes" or omitting the "--comp-lzo"
+	 * flag.
+	 *
+	 * Internally, we also support "--comp-lzo [adaptive]" and "--comp-lzo no"
+	 * which have different meaning for openvpn. */
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_COMP_LZO);
-	if (NM_IN_STRSET (value, "yes", "adaptive")) {
-		/* the UI currently only supports "--comp-lzo yes" or omitting the "--comp-lzo"
-		 * flag.
-		 *
-		 * Internally, we also support "--comp-lzo [adaptive]" and "--comp-lzo no"
-		 * which have different meaning for openvpn. */
-		widget = GTK_WIDGET (gtk_builder_get_object (builder, "lzo_checkbutton"));
-		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
-	}
+	_builder_init_toggle_button (builder, "lzo_checkbutton", NM_IN_STRSET (value, "yes", "adaptive"));
 
-	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_MSSFIX);
-	if (value && !strcmp (value, "yes")) {
-		widget = GTK_WIDGET (gtk_builder_get_object (builder, "mssfix_checkbutton"));
-		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
-	}
+	_builder_init_toggle_button (builder, "mssfix_checkbutton", _hash_get_boolean (hash, NM_OPENVPN_KEY_MSSFIX));
+	_builder_init_toggle_button (builder, "float_checkbutton", _hash_get_boolean (hash, NM_OPENVPN_KEY_FLOAT));
+	_builder_init_toggle_button (builder, "tcp_checkbutton", _hash_get_boolean (hash, NM_OPENVPN_KEY_PROTO_TCP));
 
-	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_FLOAT);
-	if (value && !strcmp (value, "yes")) {
-		widget = GTK_WIDGET (gtk_builder_get_object (builder, "float_checkbutton"));
-		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
-	}
-
-	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_PROTO_TCP);
-	if (value && !strcmp (value, "yes")) {
-		widget = GTK_WIDGET (gtk_builder_get_object (builder, "tcp_checkbutton"));
-		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
-	}
 
 	/* Populate device-related widgets */
 	dev =      g_hash_table_lookup (hash, NM_OPENVPN_KEY_DEV);
@@ -1692,60 +1793,34 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 	gtk_entry_set_placeholder_text (GTK_ENTRY (entry), _("(automatic)"));
 	g_signal_connect (G_OBJECT (entry), "insert-text", G_CALLBACK (device_name_filter_cb), NULL);
 	g_signal_connect (G_OBJECT (entry), "changed", G_CALLBACK (device_name_changed_cb), ok_button);
-	if (dev && dev[0] != '\0')
-		gtk_entry_set_text (GTK_ENTRY (entry), dev);
+	gtk_entry_set_text (GTK_ENTRY (entry), dev ?: "");
 
-	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_REMOTE_RANDOM);
-	if (value && !strcmp (value, "yes")) {
-		widget = GTK_WIDGET (gtk_builder_get_object (builder, "remote_random_checkbutton"));
-		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
-	}
 
-	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_TUN_IPV6);
-	if (value && !strcmp (value, "yes")) {
-		widget = GTK_WIDGET (gtk_builder_get_object (builder, "tun_ipv6_checkbutton"));
-		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
-	}
+	_builder_init_toggle_button (builder, "remote_random_checkbutton", _hash_get_boolean (hash, NM_OPENVPN_KEY_REMOTE_RANDOM));
+	_builder_init_toggle_button (builder, "tun_ipv6_checkbutton", _hash_get_boolean (hash, NM_OPENVPN_KEY_TUN_IPV6));
 
 	widget = GTK_WIDGET (gtk_builder_get_object (builder, "cipher_combo"));
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_CIPHER);
 	populate_cipher_combo (GTK_COMBO_BOX (widget), value);
 
-	widget = GTK_WIDGET (gtk_builder_get_object (builder, "keysize_checkbutton"));
-	g_assert (widget);
-	spin = GTK_WIDGET (gtk_builder_get_object (builder, "keysize_spinbutton"));
-	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (checkbox_toggled_update_widget_cb), spin);
 
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_KEYSIZE);
-	if (value && strlen (value)) {
-		long int tmp;
+	_builder_init_optional_spinbutton (builder, "keysize_checkbutton", "keysize_spinbutton", !!value,
+	                                   _nm_utils_ascii_str_to_int64 (value, 10, 1, 65535, 128));
 
-		errno = 0;
-		tmp = strtol (value, NULL, 10);
-		if (errno == 0 && tmp > 0 && tmp < 65536) {
-			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
-
-			widget = GTK_WIDGET (gtk_builder_get_object (builder, "keysize_spinbutton"));
-			gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), (gdouble) tmp);
-			gtk_widget_set_sensitive (widget, TRUE);
-		}
-	} else {
-		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), FALSE);
-
-		widget = GTK_WIDGET (gtk_builder_get_object (builder, "keysize_spinbutton"));
-		gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), 128.0);
-		gtk_widget_set_sensitive (widget, FALSE);
-	}
 
 	widget = GTK_WIDGET (gtk_builder_get_object (builder, "hmacauth_combo"));
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_AUTH);
 	populate_hmacauth_combo (GTK_COMBO_BOX (widget), value);
 
-	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_TLS_REMOTE);
-	if (value && strlen (value)) {
-		widget = GTK_WIDGET (gtk_builder_get_object (builder, "tls_remote_entry"));
-		gtk_entry_set_text (GTK_ENTRY(widget), value);
-	}
+	entry = GTK_WIDGET (gtk_builder_get_object (builder, "tls_remote_entry"));
+	combo = GTK_WIDGET (gtk_builder_get_object (builder, "tls_remote_mode_combo"));
+	populate_tls_remote_mode_entry_combo (GTK_ENTRY (entry), GTK_COMBO_BOX (combo),
+	                                      g_hash_table_lookup (hash, NM_OPENVPN_KEY_TLS_REMOTE),
+	                                      g_hash_table_lookup (hash, NM_OPENVPN_KEY_VERIFY_X509_NAME));
+	g_signal_connect (G_OBJECT (entry), "changed", G_CALLBACK (tls_remote_changed), builder);
+	g_signal_connect (G_OBJECT (combo), "changed", G_CALLBACK (tls_remote_changed), builder);
+	tls_remote_changed (entry, builder);
 
 	widget = GTK_WIDGET (gtk_builder_get_object (builder, "remote_cert_tls_checkbutton"));
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_REMOTE_CERT_TLS);
@@ -1820,43 +1895,24 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 		gtk_notebook_remove_page (GTK_NOTEBOOK (widget), 2);
 	}
 
-	/* ping */
-	widget = GTK_WIDGET (gtk_builder_get_object (builder, "ping_checkbutton"));
-	spin = GTK_WIDGET (gtk_builder_get_object (builder, "ping_spinbutton"));
-	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (checkbox_toggled_update_widget_cb), spin);
+
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_PING);
-	if (value && *value) {
-		long int tmp;
+	_builder_init_optional_spinbutton (builder, "ping_checkbutton", "ping_spinbutton", !!value,
+	                                   _nm_utils_ascii_str_to_int64 (value, 10, 1, 65535, 30));
 
-		errno = 0;
-		tmp = strtol (value, NULL, 10);
-		if (errno == 0 && tmp > 0 && tmp < 65536) {
-			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
-
-			widget = GTK_WIDGET (gtk_builder_get_object (builder, "ping_spinbutton"));
-			gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), (gdouble) tmp);
-			gtk_widget_set_sensitive (widget, TRUE);
-		}
-	} else {
-		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), FALSE);
-
-		widget = GTK_WIDGET (gtk_builder_get_object (builder, "ping_spinbutton"));
-		gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), 30.0);
-		gtk_widget_set_sensitive (widget, FALSE);
-	}
 
 	/* ping-exit / ping-restart */
 	widget = GTK_WIDGET (gtk_builder_get_object (builder, "ping_exit_restart_checkbutton"));
-	ping_exit_restart_checkbox_toggled_cb (widget, builder);
-	g_signal_connect (G_OBJECT (widget), "toggled", G_CALLBACK (ping_exit_restart_checkbox_toggled_cb), builder);
-
+	spin = GTK_WIDGET (gtk_builder_get_object (builder, "ping_exit_restart_spinbutton"));
 	combo = GTK_WIDGET (gtk_builder_get_object (builder, "ping_exit_restart_combo"));
+	g_signal_connect ((GObject *) widget, "toggled", G_CALLBACK (ping_exit_restart_checkbox_toggled_cb), builder);
+
 	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_PING_EXIT);
-	if (value && *value)
-		active = PING_EXIT;
-	else {
+	active = PING_EXIT;
+	if (!value) {
 		value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_PING_RESTART);
-		active = PING_RESTART;
+		if (value)
+			active = PING_RESTART;
 	}
 
 	store = gtk_list_store_new (1, G_TYPE_STRING);
@@ -1866,27 +1922,19 @@ advanced_dialog_new (GHashTable *hash, const char *contype)
 	gtk_list_store_set (store, &iter, 0, _("ping-restart"), -1);
 	gtk_combo_box_set_model (GTK_COMBO_BOX (combo), GTK_TREE_MODEL (store));
 	g_object_unref (store);
-	gtk_combo_box_set_active (GTK_COMBO_BOX (combo), active);
+	gtk_combo_box_set_active ((GtkComboBox *) combo, active);
 
-	if (value && *value) {
-		long int tmp;
+	gtk_spin_button_set_value ((GtkSpinButton *) spin,
+	                           (double) _nm_utils_ascii_str_to_int64 (value, 10, 1, 65535, 30));
+	gtk_widget_set_sensitive (combo, !!value);
+	gtk_widget_set_sensitive (spin, !!value);
+	gtk_toggle_button_set_active ((GtkToggleButton *) widget, !!value);
 
-		errno = 0;
-		tmp = strtol (value, NULL, 10);
-		if (errno == 0 && tmp > 0 && tmp < 65536) {
-			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), TRUE);
 
-			widget = GTK_WIDGET (gtk_builder_get_object (builder, "ping_exit_restart_spinbutton"));
-			gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), (gdouble) tmp);
-			gtk_widget_set_sensitive (widget, TRUE);
-		}
-	} else {
-		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (widget), FALSE);
+	value = g_hash_table_lookup (hash, NM_OPENVPN_KEY_MAX_ROUTES);
+	_builder_init_optional_spinbutton (builder, "max_routes_checkbutton", "max_routes_spinbutton", !!value,
+	                                   _nm_utils_ascii_str_to_int64 (value, 10, 0, 100000000, 100));
 
-		widget = GTK_WIDGET (gtk_builder_get_object (builder, "ping_exit_restart_spinbutton"));
-		gtk_spin_button_set_value (GTK_SPIN_BUTTON (widget), 30.0);
-		gtk_widget_set_sensitive (widget, FALSE);
-	}
 
 out:
 	g_free (ui_file);
@@ -1897,7 +1945,7 @@ GHashTable *
 advanced_dialog_new_hash_from_dialog (GtkWidget *dialog, GError **error)
 {
 	GHashTable *hash;
-	GtkWidget *widget;
+	GtkWidget *widget, *entry, *combo;
 	GtkBuilder *builder;
 	const char *contype = NULL;
 	const char *value;
@@ -2010,7 +2058,10 @@ advanced_dialog_new_hash_from_dialog (GtkWidget *dialog, GError **error)
 	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
 		/* we only have a checkbox, which we either map to "--comp-lzo yes" or
 		 * no "--comp-lzo" flag. In the UI, we cannot express "--comp-lzo [adaptive]"
-		 * or "--comp-lzo no". */
+		 * or "--comp-lzo no".
+		 *
+		 * Note that "--comp-lzo no" must be encoded as "comp-lzo=no-by-default" (bgo#769177).
+		 */
 		g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_COMP_LZO), g_strdup ("yes"));
 	}
 
@@ -2092,10 +2143,25 @@ advanced_dialog_new_hash_from_dialog (GtkWidget *dialog, GError **error)
 	    || !strcmp (contype, NM_OPENVPN_CONTYPE_PASSWORD_TLS)
 	    || !strcmp (contype, NM_OPENVPN_CONTYPE_PASSWORD)) {
 
-		widget = GTK_WIDGET (gtk_builder_get_object (builder, "tls_remote_entry"));
-		value = gtk_entry_get_text (GTK_ENTRY(widget));
-		if (value && strlen (value))
-			g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_TLS_REMOTE), g_strdup (value));
+		entry = GTK_WIDGET (gtk_builder_get_object (builder, "tls_remote_entry"));
+		value = gtk_entry_get_text (GTK_ENTRY (entry));
+
+		combo = GTK_WIDGET (gtk_builder_get_object (builder, "tls_remote_mode_combo"));
+		model = gtk_combo_box_get_model (GTK_COMBO_BOX (combo));
+
+		if (value && strlen (value) && gtk_combo_box_get_active_iter (GTK_COMBO_BOX (combo), &iter)) {
+			gs_free char *tls_remote_mode = NULL;
+			gtk_tree_model_get (model, &iter, TLS_REMOTE_MODE_COL_VALUE, &tls_remote_mode, -1);
+
+			if (!g_strcmp0 (tls_remote_mode, TLS_REMOTE_MODE_NONE)) {
+				// pass
+			} else if (!g_strcmp0 (tls_remote_mode, TLS_REMOTE_MODE_LEGACY)) {
+				g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_TLS_REMOTE), g_strdup(value));
+			} else {
+				char *x509_name = g_strdup_printf ("%s:%s", tls_remote_mode, value);
+				g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_VERIFY_X509_NAME), x509_name);
+			}
+		}
 
 		widget = GTK_WIDGET (gtk_builder_get_object (builder, "remote_cert_tls_checkbutton"));
 		if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
@@ -2152,7 +2218,7 @@ advanced_dialog_new_hash_from_dialog (GtkWidget *dialog, GError **error)
 		}
 	}
 
-	widget = GTK_WIDGET (gtk_builder_get_object (builder, "ping_checkbutton"));
+        widget = GTK_WIDGET (gtk_builder_get_object (builder, "ping_checkbutton"));
 	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
 		int ping_val;
 
@@ -2181,6 +2247,16 @@ advanced_dialog_new_hash_from_dialog (GtkWidget *dialog, GError **error)
 		                     g_strdup_printf ("%d", ping_val));
 	}
 
+	/* max routes */
+	widget = GTK_WIDGET (gtk_builder_get_object (builder, "max_routes_checkbutton"));
+	if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget))) {
+		int max_routes;
+
+		widget = GTK_WIDGET (gtk_builder_get_object (builder, "max_routes_spinbutton"));
+		max_routes = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (widget));
+		g_hash_table_insert (hash, g_strdup (NM_OPENVPN_KEY_MAX_ROUTES), g_strdup_printf ("%d", max_routes));
+	}
+
 	return hash;
 }
 
